(********************************************************************)
(* SessML - Runtime                                                 *)
(*                                                                  *)
(* amqp.ml                                                          *)
(********************************************************************)
(* Time-stamp: <Pierre-Malo Denielou - 2011> *)

open Unix
open Global
open Data
open Prins


(**************************************)
(* AMQP messages codes and parameters *)
(**************************************)

  
type param = 
    Param_octet | Param_short | Param_long | Param_longlong | Param_bool | Param_bool2 |
        Param_bool3 |Param_bool4 | Param_bool5 | Param_shortstr | Param_longstr | Param_table


type pval = 
  | Pval_bool of bool
  | Pval_bool2 of (bool * bool)
  | Pval_bool3 of (bool * bool * bool)
  | Pval_bool4 of (bool * bool * bool * bool)
  | Pval_bool5 of (bool * bool * bool * bool * bool)
  | Pval_octet of int
  | Pval_short of int
  | Pval_long of int
  | Pval_longlong of int
  | Pval_shortstr of string
  | Pval_longstr of string
  | Pval_table of (string * pval) list


let rec sprint_amqp_pval = function
  | Pval_bool b -> string_of_bool b
  | Pval_bool2 (a,b) -> 
      string_of_bool a ^","^string_of_bool b
  | Pval_bool3 (a,b,c) -> 
      string_of_bool a ^","^string_of_bool b^","^string_of_bool c
  | Pval_bool4 (a,b,c,d) -> 
      string_of_bool a ^","^string_of_bool b ^","^string_of_bool c ^","^string_of_bool d
  | Pval_bool5 (a,b,c,d,e) -> 
      string_of_bool a ^","^string_of_bool b ^","^string_of_bool c
      ^","^string_of_bool d ^","^string_of_bool e
  | Pval_octet i -> string_of_int i
  | Pval_short i  -> string_of_int i
  | Pval_long i  -> string_of_int i
  | Pval_longlong i  -> string_of_int i
  | Pval_shortstr s -> "\""^s^"\""
  | Pval_longstr s -> "\""^s^"\""
  | Pval_table t -> 
      "["^(String.concat ";" (List.map (function (k,v)-> k^":"^sprint_amqp_pval v) t))^"]"
    
let sprint_amqp_param l = String.concat ";" (List.map sprint_amqp_pval l)

let id_to_string = function
  | 10 -> 
      (function 
         | 10 -> "Connection","Start"
         | 11 -> "Connection","Start_Ok"
         | 20 -> "Connection","Secure"
         | 21 -> "Connection","Secure_Ok"
         | 30 -> "Connection","Tune"
         | 31 -> "Connection","Tune_Ok"
         | 40 -> "Connection","Open"
         | 41 -> "Connection","Open_Ok"
        (* | 50 -> "Connection","Redirect"*)
         | 50 -> "Connection","Close"
         | 51 -> "Connection","Close_Ok"
         | _ -> "Connection","Autre")
  | 20 -> 
      (function 
         | 10 -> "Channel","Open"
         | 11 -> "Channel","Open_Ok"
         | 20 -> "Channel","Flow"
         | 21 -> "Channel","Flow_Ok"
         | 30 -> "Channel","Alert"
         | 40 -> "Channel","Close"
         | 41 -> "Channel","Close_Ok"
         | _ -> "Channel","Autre")
  | 30 -> 
      (function 
         | 10 -> "Access","Request"
         | 11 -> "Access","Request_Ok"
         | _ -> "Access","Autre")
  | 40 -> 
      (function
         | 10 -> "Exchange","Declare"
         | 11 -> "Exchange","Declare_Ok"
         | 20 -> "Exchange","Delete"
         | 21 -> "Exchange","Delete_Ok"
         | _ -> "Autre","Exchange")
  | 50 -> 
      (function
         | 10 -> "Queue","Declare"
         | 11 -> "Queue","Declare_Ok"
         | 20 -> "Queue","Bind"
         | 21 -> "Queue","Bind_Ok"
         | 40 -> "Queue","Delete"
         | 41 -> "Queue","Delete_Ok"
         | _ -> "Autre","Queue")
  | 60 -> 
      (function
         | 20 -> "Basic","Consume"
         | 21 -> "Basic","Consume_Ok"
         | 40 -> "Basic","Publish"
         | 60 -> "Basic","Deliver"
         | 80 -> "Basic","Ack"
         | _ -> "Autre","Basic")
  | _ -> 
      (function
         | _ -> "Autre","Autre")


let string_to_id = function
  | "Connection","Start"    -> 10,10     
  | "Connection","Start_Ok" -> 10,11      
  | "Connection","Secure"   -> 10,20    
  | "Connection","Secure_Ok"-> 10,21    
  | "Connection","Tune"     -> 10,30   
  | "Connection","Tune_Ok"  -> 10,31  
  | "Connection","Open"     -> 10,40
  | "Connection","Open_Ok"  -> 10,41      
  | "Connection","Close"    -> 10,50    
  | "Connection","Close_Ok" -> 10,51   
  | "Channel","Open"     -> 20,10    
  | "Channel","Open_Ok"  -> 20,11       
  | "Channel","Flow"     -> 20,20    
  | "Channel","Flow_Ok"  -> 20,21       
  | "Channel","Alert"    -> 20,30     
  | "Channel","Close"    -> 20,40 
  | "Channel","Close_Ok" -> 20,41   
  | "Exchange","Declare" -> 40,10  
  | "Exchange","Declare_Ok" -> 40,11 
  | "Exchange","Delete" -> 40,20  
  | "Exchange","Delete_Ok" -> 40,21
  | "Queue","Declare"    -> 50,10  
  | "Queue","Declare_Ok" -> 50,11
  | "Queue","Bind"       -> 50,20  
  | "Queue","Bind_Ok"    -> 50,21
  | "Queue","Delete"     -> 50,40  
  | "Queue","Delete_Ok"  -> 50,41
  | "Access","Request"   -> 0,10
  | "Access","Request_Ok"-> 0,11  
  | "Basic","Consume"    -> 60,20
  | "Basic","Consume_Ok" -> 60,21
  | "Basic","Publish"    -> 60,40
  | "Basic","Deliver"    -> 60,60
  | "Basic","Ack"        -> 60,80
  | _ -> assert false


let string_to_param = function
  | "Connection","Start"    -> [Param_octet;Param_octet;Param_table;Param_longstr;Param_longstr]
  | "Connection","Start_Ok" -> [Param_table;Param_shortstr;Param_longstr;Param_shortstr]
  | "Connection","Tune"     -> [Param_short;Param_long;Param_short]
  | "Connection","Tune_Ok"  -> [Param_short;Param_long;Param_short]
  | "Connection","Open"     -> [Param_shortstr;Param_shortstr;Param_bool]
  | "Connection","Open_Ok"  -> [Param_shortstr]
  | "Connection","Close"    -> [Param_short;Param_shortstr;Param_short;Param_short]
  | "Connection","Close_Ok" -> []
  | "Channel","Open"        -> [Param_shortstr]
  | "Channel","Open_Ok"     -> []
  | "Channel","Close"       -> [Param_short;Param_shortstr;Param_short;Param_short]
  | "Channel","Close_Ok"    -> []
  | "Queue","Declare"       -> [Param_short;Param_shortstr;Param_bool5;Param_table]
  | "Queue","Declare_Ok"    -> [Param_shortstr;Param_long;Param_long]
  | "Queue","Bind"          -> [Param_short;Param_shortstr;Param_shortstr;Param_shortstr;Param_bool;Param_table]
  | "Queue","Bind_Ok"       -> []
  | "Queue","Delete"        -> [Param_short;Param_shortstr;Param_bool3]
  | "Queue","Delete_Ok"     -> [Param_long]
  | "Exchange","Declare"    -> [Param_short;Param_shortstr;Param_shortstr;Param_bool5;Param_table]
  | "Exchange","Declare_Ok" -> []
  | "Exchange","Delete"     -> [Param_short;Param_shortstr;Param_bool2]
  | "Exchange","Delete_Ok"  -> []
  | "Basic","Consume"       -> [Param_short;Param_shortstr;Param_shortstr;Param_bool4;Param_table]
  | "Basic","Consume_Ok"    -> [Param_shortstr]
  | "Basic","Publish"       -> [Param_short;Param_shortstr;Param_shortstr;Param_bool2]
  | "Basic","Deliver"       -> [Param_shortstr;Param_longlong;Param_bool;Param_shortstr;Param_shortstr]
  | "Basic","Ack"           -> [Param_longlong;Param_bool]
  | _ -> assert false

(**************)
(* Exceptions *)
(**************)

type error =
  | Content_too_large
  | No_consumers
  | Connection_forced
  | Invalid_path
  | Access_refused
  | Notfound
  | Resource_locked
  | Precondition_failed
  | Frame_error
  | Syntax_error
  | Command_invalid
  | Channel_error
  | Unexpected_frame
  | Resource_error
  | Not_allowed
  | Not_implemented
  | Internal_error
  | Unknown_error

exception Error of (error * string * string * string)

let amqp_exception code errmsg cla meth = 
  let cla,met = id_to_string cla meth in
  let error = match code with
    | 311 -> Content_too_large
    | 313 -> No_consumers
    | 320 -> Connection_forced
    | 402 -> Invalid_path
    | 403 -> Access_refused
    | 404 -> Notfound
    | 405 -> Resource_locked
    | 406 -> Precondition_failed
    | 501 -> Frame_error
    | 502 -> Syntax_error
    | 503 -> Command_invalid
    | 504 -> Channel_error
    | 505 -> Unexpected_frame
    | 506 -> Resource_error
    | 530 -> Not_allowed
    | 540 -> Not_implemented
    | 541 -> Internal_error
    | _ -> Unknown_error
  in
  raise (Error (error,errmsg,cla,met))
      
exception Connection_stop
  

(*************)
(* Encodings *)
(*************)

let frame_end = '\206'

let code = Char.code
let chr = Char.chr
let to_char = chr

let to_bool2 = function
  | false,false -> 0
  | false,true -> 2
  | true,false -> 1
  | true,true -> 3

let to_bool3 = function
  | false,false,false -> 0
  | false,false,true -> 2
  | false,true,false -> 1
  | false,true,true -> 3
  | true,false,false -> 4
  | true,false,true -> 6
  | true,true,false -> 5
  | true,true,true -> 7

let to_bool4 = function
  | false,false,false,false -> 0 
  | false,false,false,true ->  8
  | false,false,true,false ->  4
  | false,false,true,true ->   12
  | false,true,false,false ->  2
  | false,true,false,true ->   10
  | false,true,true,false ->   6
  | false,true,true,true ->    14
  | true,false,false,false ->  1
  | true,false,false,true ->   9
  | true,false,true,false ->   5
  | true,false,true,true ->    13
  | true,true,false,false ->   3
  | true,true,false,true ->    11
  | true,true,true,false ->    7
  | true,true,true,true ->     15


let to_bool5 = function
  | false,false,false,false,false -> 0 
  | false,false,false,true,false ->  8
  | false,false,true,false,false ->  4
  | false,false,true,true,false ->   12
  | false,true,false,false,false ->  2
  | false,true,false,true,false ->   10
  | false,true,true,false,false ->   6
  | false,true,true,true,false ->    14
  | true,false,false,false,false ->  1
  | true,false,false,true,false ->   9
  | true,false,true,false,false ->   5
  | true,false,true,true,false ->    13
  | true,true,false,false,false ->   3
  | true,true,false,true,false ->    11
  | true,true,true,false,false ->    7
  | true,true,true,true,false ->     15
  | false,false,false,false,true -> 16 
  | false,false,false,true,true ->  24
  | false,false,true,false,true ->  20
  | false,false,true,true,true ->   28
  | false,true,false,false,true ->  18
  | false,true,false,true,true ->   26
  | false,true,true,false,true ->   22
  | false,true,true,true,true ->    30
  | true,false,false,false,true ->  17
  | true,false,false,true,true ->   25
  | true,false,true,false,true ->   21
  | true,false,true,true,true ->    29
  | true,true,false,false,true ->   19
  | true,true,false,true,true ->    27
  | true,true,true,false,true ->    23
  | true,true,true,true,true ->     31

let to_short n = 
  Printf.sprintf "%c%c" (chr (n/256)) (chr (n mod 256))

let to_long n = Printf.sprintf "%c%c%c%c" 
  (chr (n/(256*256*256))) 
  (chr ((n/(256*256)) mod 256))
  (chr ((n/256) mod 256))
  (chr (n mod 256))

let to_longlong n = Printf.sprintf "\000\000\000\000%c%c%c%c" 
  (chr (n/(256*256*256))) 
  (chr ((n/(256*256)) mod 256))
  (chr ((n/256) mod 256))
  (chr (n mod 256))


let to_shortstring s =
  let n = String.length s in 
  Printf.sprintf "%c%s" (chr n) s

let to_longstring s = 
  let n = String.length s in 
  Printf.sprintf "%s%s" (to_long n) s
  
let to_field_longstring s = "S"^to_longstring s

let to_field = function
  | Pval_longstr s -> "S"^to_longstring s
  | _ -> assert false
    
let to_table (l:(string*pval) list) = 
  let ll = List.map (function (n,v) -> (to_shortstring n)^ to_field v) l in
  let t = String.concat "" ll in
  let len = String.length t in
  (to_long len)^t

  
  
(********************)
(* Message handling *)
(********************)

let rec build_amqp_param = function
    [] -> ""
  | Pval_bool true ::q ->  "\001"^(build_amqp_param q)
  | Pval_bool false ::q -> "\000"^(build_amqp_param q)
  | Pval_bool2 b ::q -> (String.make 1 (to_char (to_bool2 b)))^(build_amqp_param q)
  | Pval_bool3 b ::q -> (String.make 1 (to_char (to_bool3 b)))^(build_amqp_param q)
  | Pval_bool4 b ::q -> (String.make 1 (to_char (to_bool4 b)))^(build_amqp_param q)
  | Pval_bool5 b ::q -> (String.make 1 (to_char (to_bool5 b)))^(build_amqp_param q)
  | Pval_octet i ::q -> String.make 1 (to_char i) ^(build_amqp_param q)
  | Pval_short i ::q  -> to_short i ^(build_amqp_param q)
  | Pval_long i ::q -> to_long i ^(build_amqp_param q)
  | Pval_longlong i ::q -> to_longlong i ^(build_amqp_param q)
  | Pval_shortstr s ::q -> to_shortstring s ^(build_amqp_param q)
  | Pval_longstr s ::q -> to_longstring s ^(build_amqp_param q)
  | Pval_table t ::q -> to_table t ^(build_amqp_param q)

let rec read_amqp_table n s =
  if s = "" then []
  else begin
    let len = code s.[0] in
    let key = String.sub s 1 len in
    let typ = s.[len+1] in
    match typ with
      | 'S' -> let leng = Scanf.sscanf (String.sub s (len+2) 4) "%c%c%c%c"
          (fun l1 l2 l3 l4 -> 
             code l1 *256*256*256 + code l2 *256*256
             + code l3 *256+ code l4) in 
        (key, Pval_longstr (String.sub s (len+6) leng))::(read_amqp_table (n-(len+6+leng))
                                             (String.sub s (len+6+leng) (n-(len+6+leng))))
      | _ -> assert false
  end
          
let rec read_amqp_param n s = function
    [] -> []
  | Param_bool  ::q -> let () = debug_com_full (Printf.sprintf "Reading boolean") in
    (if code s.[0] = 1 then (Pval_bool true) else (Pval_bool false))
    ::(read_amqp_param (n-1) (String.sub s 1 (n-1)) q)
  | Param_bool2  ::q -> let () = debug_com_full (Printf.sprintf "Reading boolean2") in
    (let c = code s.[0] in
     Pval_bool2 ((c mod 2=1),(c/2=1)))
    ::(read_amqp_param (n-1) (String.sub s 1 (n-1)) q)
  | Param_bool3  ::q -> let () = debug_com_full (Printf.sprintf "Reading boolean3") in
    (let c = code s.[0] in
     Pval_bool3 ((c mod 2=1),(c/2 mod 2=1),(c/4 =1)))
    ::(read_amqp_param (n-1) (String.sub s 1 (n-1)) q)
  | Param_bool4  ::q -> let () = debug_com_full (Printf.sprintf "Reading boolean4") in
    (let c = code s.[0] in
     Pval_bool4 ((c mod 2=1),(c/2 mod 2=1),(c/4 mod 2=1),(c/8 = 1)))
    ::(read_amqp_param (n-1) (String.sub s 1 (n-1)) q)
  | Param_bool5  ::q -> let () = debug_com_full (Printf.sprintf "Reading boolean5") in
    (let c = code s.[0] in
     Pval_bool5 ((c mod 2=1),(c/2 mod 2=1),(c/4 mod 2=1),(c/8 mod 2= 1),(c/16 = 1)))
    ::(read_amqp_param (n-1) (String.sub s 1 (n-1)) q)
  | Param_octet  ::q -> let () = debug_com_full (Printf.sprintf "Reading octet") in
    (Pval_octet (code s.[0]))
    ::(read_amqp_param (n-1) (String.sub s 1 (n-1)) q)
  | Param_short ::q -> let () = debug_com_full (Printf.sprintf "Reading short") in
    (Pval_short (Scanf.sscanf (String.sub s 0 2) "%c%c"
                   (fun a b -> code a * 256 + code b)))
    ::(read_amqp_param (n-2) (String.sub s 2 (n-2)) q)
  | Param_long ::q -> let () = debug_com_full (Printf.sprintf "Reading long") in
    (Pval_long (Scanf.sscanf (String.sub s 0 4) "%c%c%c%c"
                  (fun l1 l2 l3 l4 -> 
                     code l1 *256*256*256 + code l2 *256*256
                     + code l3 *256+ code l4)))
    ::(read_amqp_param (n-4) (String.sub s 4 (n-4)) q)
  | Param_longlong ::q -> let () = debug_com_full (Printf.sprintf "Reading longlong") in
    (Pval_longlong (Scanf.sscanf (String.sub s 0 8) "\000\000\000\000%c%c%c%c"
                  (fun l1 l2 l3 l4 -> 
                     code l1 *256*256*256 + code l2 *256*256
                     + code l3 *256+ code l4)))
    ::(read_amqp_param (n-8) (String.sub s 4 (n-8)) q)
  | Param_shortstr ::q -> let () = debug_com_full (Printf.sprintf "Reading short string") in
    let len = code s.[0] in 
    (Pval_shortstr (String.sub s 1 len))
    ::(read_amqp_param (n-(len+1)) (String.sub s (len+1) (n-(len+1))) q)
  | Param_longstr ::q -> let () = debug_com_full (Printf.sprintf "Reading long string") in
    let len = Scanf.sscanf (String.sub s 0 4) "%c%c%c%c"
      (fun l1 l2 l3 l4 -> 
         code l1 *256*256*256 + code l2 *256*256
         + code l3 *256+ code l4) in 
    (Pval_longstr (String.sub s 4 len))
    ::(read_amqp_param (n-(len+4)) (String.sub s (len+4) (n-(len+4))) q)
  | Param_table ::q -> let () = debug_com_full (Printf.sprintf "Reading table") in
    let len = Scanf.sscanf (String.sub s 0 4) "%c%c%c%c"
      (fun l1 l2 l3 l4 -> 
         code l1 *256*256*256 + code l2 *256*256
         + code l3 *256+ code l4) in 
    (Pval_table (read_amqp_table len (String.sub s 4 len)))
    ::(read_amqp_param (n-(len+4)) (String.sub s (len+4) (n-(len+4))) q)


let build_amqp_method_message chan cla met args =
  let cla,met = string_to_id (cla,met) in
  let frame = Printf.sprintf "%s%s%s" (to_short cla) (to_short met) (build_amqp_param args) in
  let size = String.length frame in 
  let msg = Printf.sprintf "%c%s%s%s%c"
    '\001' (to_short chan) (to_long size) frame frame_end
  in
  msg

let build_amqp_content_header_message chan class_id ?(weight=0) length =
  let frame = Printf.sprintf "%s%s%s%s\x90\x00%s" (to_short class_id) (to_short weight) (to_long 0)
    (to_long length)  
    (build_amqp_param [Pval_shortstr "text/plain"; Pval_octet 2]) in
  let size = String.length frame in 
  let msg = Printf.sprintf "%c%s%s%s%c"
    '\002' (to_short chan) (to_long size) frame frame_end
  in
  msg

let build_amqp_content_message chan payload =
  let size = String.length payload in 
  let msg = Printf.sprintf "%c%s%s%s%c"
    '\003' (to_short chan) (to_long size) payload frame_end
  in
  msg

type amqp_message = 
    Method of (string*string)* (pval list)
  | CHeader of (int)
  | CBody of (string)

let read_amqp_message msg =
  let head = String.sub msg 0 7 in
  let (a,c1,c2,l1,l2,l3,l4) = Scanf.sscanf head "%c%c%c%c%c%c%c" 
    (fun a -> fun c1 -> fun c2 -> fun l1 -> fun l2 -> fun l3 -> fun l4 ->
       (a,c1,c2,l1,l2,l3,l4)) in
  let chan = code c1 * 256 + code c2 in
  let length = code l1 *256*256*256 + code l2 *256*256+ code l3 *256+ code l4
  in
  let payload = String.sub msg 7 length in
  (* let _ = String.sub msg (7+length) 1 in *)
  if code a = 1 then begin
    let payload_header = String.sub payload 0 4 in
    let payload_args = String.sub payload 4 (length-4) in
    let (cid1,cid2,mid1,mid2)=
      Scanf.sscanf payload_header "%c%c%c%c" 
        (fun cid1 cid2 mid1 mid2 -> (cid1,cid2,mid1,mid2)) in
    let class_id = code cid1 *256 + code cid2 in
    let meth_id = code mid1 *256 + code mid2 in
    let cla,met = id_to_string class_id meth_id in
    let param = string_to_param (cla,met) in
    (*  debug_com (Printf.sprintf 
        "AMQP message:\n-kind: %d  -channel: %d  -length: %d  -class/meth: %s.%s\n args: %s\n end: %s\n" 
        
        (code a) chan length cla met payload_args payload_end);*)
    let vparam = read_amqp_param (length-4) payload_args param in
    let sparam = sprint_amqp_param vparam in
    debug_com (Printf.sprintf 
                 "AMQP method message:\n kind: %d, channel: %d, length: %d, class/meth: %s.%s\n args: %s" 
                 
                 (code a) chan length cla met sparam);
    Method ((cla,met),vparam)
  end
  else if code a = 2 then begin
    let payload_header = String.sub payload 0 4 in
    let payload_args = String.sub payload 4 (length-4) in
    let (cid1,cid2,mid1,mid2)=
      Scanf.sscanf payload_header "%c%c%c%c" 
        (fun cid1 cid2 mid1 mid2 -> (cid1,cid2,mid1,mid2)) in
    let class_id = code cid1 *256 + code cid2 in
    let _ (* weight *) = code mid1 *256 + code mid2 in
    let head = String.sub payload_args 0 10 in
    let (_,_,_,_,l5,l6,l7,l8,_,_) = Scanf.sscanf head "%c%c%c%c%c%c%c%c%c%c" 
      (fun l1 -> fun l2 -> fun l3 -> fun l4 -> fun l5 -> fun l6 -> fun l7 -> fun
         l8 -> fun c1 -> fun c2 ->
           (l1,l2,l3,l4,l5,l6,l7,l8,c1,c2)) in
    let paylength = code l5 *256*256*256 + code l6 *256*256+ code l7 *256+ code l8
    in
(*    let payload_args = String.sub payload 14 (length-14) in
    debug_com (Printf.sprintf 
                 "Parsing a content header AMQP message payload of length %d:
    %S" (length-14) payload_args);
    let vparam = read_amqp_param (length-14) payload_args [Param_shortstr; Param_octet] in
    debug_com (Printf.sprintf 
                 "Parsing a content header AMQP message 5");
    let sparam = sprint_amqp_param vparam in *)
    let sparam = "" in
    debug_com (Printf.sprintf 
                 "AMQP content header:\n kind: %d, channel: %d, length: %d, class: %d\n args: %s" 
                 
                 (code a) chan length class_id sparam);
    CHeader(paylength)
  end
  else if code a =3 then begin
    debug_com (Printf.sprintf 
                 "AMQP content:\n kind: %d, channel: %d, length: %d, payload: %s"
                 (code a) chan length payload);
    CBody(payload)
  end
  else assert false
    

(**************)
(* AMQP state *)
(**************)

let out_amqp = ref Pervasives.stdout
let in_amqp = ref stdin

(***********************************)
(* Main Send and Receive functions *)
(***********************************)

let split_amqp s =
  let rec aux i =
    if i = String.length s then
     []
    else
      try 
        let n = String.index_from s i '\206' in
        let s = String.sub s i (n-i+1) in
(*        debug_com_io ("Finding a chunck: "^s);*)
        if s="" then aux (n+1) else s::(aux (n+1))
      with Not_found -> assert false 
        (* let s = String.sub s i (String.length s -i) in
           if s = "" then [] else [s] *)
  in aux 0


let send_amqp ?(out_amqp=out_amqp) s = 
  debug_com_io (Printf.sprintf "Sending message:\n%S" s);
  let _ = try ignore (read_amqp_message s) with _ -> () in
  output_string !out_amqp s;
  flush !out_amqp 

let messages = Queue.create ()

(* Non-blocking *)
let try_recv_amqp ?(in_amqp=in_amqp) ?(messages=messages) () =
  if Queue.is_empty messages
  then begin  
    let lr,_,_ =  Unix.select [!in_amqp] [] [] 0.2 in
    match lr with
        [] -> None
      | s::_ -> begin
          let text = String.create 2056 in
          let len = recv !in_amqp text 0 2056 [] in
          if len = 0 then None (* failwith "Closing connection" *)
          else
            let msg = String.sub text 0 len in 
            debug_com_io (Printf.sprintf "A message is received of length %d" len);
            debug_com_io (Printf.sprintf "Its content is:\n%S" msg);
            let lm = split_amqp msg in
            debug_com_io (Printf.sprintf "%d parts recognized" (List.length lm));
            match lm with
                [] -> None
              | m::q -> (List.iter (fun x -> Queue.push x messages) q;Some m)
        end
  end
  else (
    debug_com_io (Printf.sprintf "Receiving a message from the queue") ;
    Some (Queue.pop messages)
  )

(* Blocking *)
let recv_amqp ?(in_amqp=in_amqp) ?(messages=messages) ?(stop=(fun ()->false))() = 
  let rec aux () = 
    match (try_recv_amqp ~in_amqp:in_amqp ~messages:messages (),stop ()) with
        None,false -> aux ()
      | None,true -> raise Connection_stop
      | Some m,_ -> m
  in
  aux ()
    
(*  if Queue.is_empty messages
  then begin
    let text = String.create 2056 in
    let len = input !in_amqp text 0 2056 in
    let msg = String.sub text 0 len in 
    debug_com_io (Printf.sprintf "A message is received of length %d" len);
    debug_com_io (Printf.sprintf "Its content is:\n%s" msg);
    let lm = split_amqp msg in
    match lm with
        [] -> assert false
      | m::q -> (List.iter (fun x -> Queue.push x messages) q;m)
  end
  else
    Queue.pop messages
*)

(* Blocking *)
let recv_amqp_constrained ?(in_amqp=in_amqp) ?(messages=messages) 
    ?(stop=(fun ()->false)) lm =
  let msg = recv_amqp ~in_amqp:in_amqp ~messages:messages ~stop:stop() in
  match read_amqp_message msg with
    | Method ((c,m),vparam) when (List.mem (c,m) lm) -> 
        (debug_com_io (Printf.sprintf "Message recognised");
         (msg,((c,m),vparam)))
    | Method (("Connection","Close"),[Pval_short code;Pval_shortstr errmsg;Pval_short
                         cla;Pval_short meth])
    | Method (("Channel","Close"),[Pval_short code;Pval_shortstr errmsg;Pval_short
                         cla;Pval_short meth]) -> 
        amqp_exception code errmsg cla meth
    | Method ((c,m),vparam) ->
      failwith (Printf.sprintf "Unexpected message %s.%s" c m)
    | _ -> assert false
    
(* Non-blocking *)
let try_recv_amqp_constrained ?(in_amqp=in_amqp) ?(messages=messages) lm =
  let msg = try_recv_amqp ~in_amqp:in_amqp ~messages:messages () in
  match msg with
      None -> None
    | Some msg -> 
        (match read_amqp_message msg with
          | Method ((c,m),vparam) when (List.mem (c,m) lm)-> (
            Some (msg,((c,m),vparam)))
          | Method ((c,m),vparam) ->
            failwith (Printf.sprintf "Unexpected message %s.%s" c m)
          | _ -> assert false
        )
    

(*************************)
(* AMQP protocol actions *)
(*************************)

(* Protocol header *)
let send_hello () =
  let hello = Printf.sprintf "AMQP\000\000\009\001" in
  send_amqp hello
  

(* Connection.Start_Ok *)
let send_Connection_Start_Ok ?(usergroup="guest") ?(username="guest")
    ?(locales="en_US") () =
  let msg = 
    build_amqp_method_message 0 "Connection" "Start_Ok" 
      [Pval_table ["product",Pval_longstr "Ocaml AMQP Client Library"];
       Pval_shortstr "PLAIN";
       Pval_longstr (Printf.sprintf "%c%s%c%s" '\000' usergroup '\000' username);
       Pval_shortstr locales] in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg

(* Connection.Tune_Ok *)
let send_Connection_Tune_Ok ?(chanmax=0) ?(framemax=10000) ?(heartbeat=0) () =
  let msg =
    build_amqp_method_message 0 "Connection" "Tune_Ok" [Pval_short chanmax;Pval_long
                                                   framemax;Pval_short heartbeat]
  in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg

(* Connection.Open *)
let send_Connection_Open ?(path="/") ?(capabilities="") ?(insist=true) () =
  let msg = build_amqp_method_message 0 "Connection" "Open"
    [Pval_shortstr path;Pval_shortstr capabilities;Pval_bool insist]
  in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg


(* Connection.Close *)
let send_Connection_Close () =
  let msg = build_amqp_method_message 0 "Connection" "Close" 
    [Pval_short 0;Pval_shortstr "";Pval_short 0 ;Pval_short 0] in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg
  
(* Channel.Open *)
let send_Channel_Open ?(channel=1) () =
  let msg = build_amqp_method_message channel "Channel" "Open" [Pval_shortstr ""] in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg
  
(* Queue.Declare *)
let send_Queue_Declare ?(channel=1) ?(ticket=1) ~queue ?(passive=false)
    ?(durable=true) ?(exclusive=false) ?(auto_delete=false) ?(no_wait=false) ()
    =  
  let msg = build_amqp_method_message channel "Queue" "Declare"
    [ Pval_short ticket; Pval_shortstr queue; 
      Pval_bool5 (passive, durable, exclusive, auto_delete,no_wait);
      Pval_table [] ] in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg
  
(* Queue.Bind *)
let send_Queue_Bind ?(channel=1) ?(ticket=1) ~queue ~exchange ~routing_key 
    ?(no_wait=false) ()
    =  
  let msg = build_amqp_method_message channel "Queue" "Bind"
    [ Pval_short ticket; Pval_shortstr queue;  Pval_shortstr exchange;
      Pval_shortstr routing_key; Pval_bool no_wait; Pval_table [] ] in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg

(* Exchange.Declare *)
let send_Exchange_Declare ?(channel=1) ?(ticket=1) ~exchange ?(extype="direct")
    ?(passive=false) ?(durable=false) ?(auto_delete=false) ?(internal=false) ?(no_wait=false) ()
    =  
  let msg = build_amqp_method_message channel "Exchange" "Declare"
    [ Pval_short ticket; Pval_shortstr exchange; Pval_shortstr extype;
      Pval_bool5 (passive, durable, auto_delete, internal, no_wait);
      Pval_table [] ] in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg
  
(* Exchange.Delete *)
let send_Exchange_Delete ?(channel=1) ?(ticket=1) ~exchange ?(if_unused=true)
    ?(no_wait=false) ()
    =  
  let msg = build_amqp_method_message channel "Exchange" "Delete"
    [ Pval_short ticket; Pval_shortstr exchange; Pval_bool2 (if_unused, no_wait)] in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg
  

(* Basic.Publish *)
let send_Basic_Publish ?(channel=1) ?(ticket=1) ~exchange ~routing
    ?(mandatory=false) ?(immediate=false) ~payload () = 
  let msg = build_amqp_method_message channel "Basic" "Publish"
    [Pval_short ticket; Pval_shortstr exchange; Pval_shortstr routing; 
     Pval_bool2 (mandatory,immediate)] in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg;
  let length = String.length payload in
  let msg = (build_amqp_content_header_message channel 60 length) in
  (* let _ = read_amqp_message msg in  *)
  send_amqp msg;
  let msg = build_amqp_content_message channel payload in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg;
  ()
    

(* Basic.Consume *)
let send_Basic_Consume ?(channel=1) ?(ticket=1) ~queue ~ctag 
    ?(no_local=false) ?(no_ack=false) ?(exclusive=false) ?(nowait=false) () =
  let msg = build_amqp_method_message channel "Basic" "Consume"
    [Pval_short ticket; Pval_shortstr queue; Pval_shortstr ctag;
     Pval_bool4(no_local,no_ack,exclusive,nowait);Pval_table []] in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg
  
(* Basic.Ack *)
let send_Basic_Ack ?(channel=1) ~delivery_tag ?(multiple=false) () =
  let msg = build_amqp_method_message channel "Basic" "Ack"
    [Pval_longlong delivery_tag;Pval_bool multiple] in
  (* let _ = read_amqp_message msg in *)
  send_amqp msg


(******************)
(* Message queues *)
(******************)

let recvqueue = Queue.create ()

let listening = ref true
let listening_thread = ref None

(*****************)
(* Communication *)
(*****************)

let open_tcp_broker ?(address="127.0.0.1") ?(port=5672) () =
  let s = socket PF_INET SOCK_STREAM 0 in
  let broker_address = inet_addr_of_string address in
  connect s (ADDR_INET (broker_address,port));
  let outc = out_channel_of_descr s in
  (outc,s)
  

(*************)
(* Interface *)
(*************)

let bind ?(interactive=true) prin =
  debug_com "Creation of the connection to the AMQP broker";
  let (outc,s) = open_tcp_broker ~port:(!amqp_port) () in
  out_amqp := outc ;
  in_amqp := s ;
  send_hello ();
  let _ = recv_amqp_constrained ["Connection","Start"] in
  send_Connection_Start_Ok ();
  let _,((cl,me),vparam) = recv_amqp_constrained ["Connection","Tune"] in
  let (_,f,_) = match vparam with 
      [Pval_short c;Pval_long f;Pval_short h]-> (c,f,h)
    | _ -> assert false in
  send_Connection_Tune_Ok ~framemax:f ();
  send_Connection_Open ();
  let _ = recv_amqp_constrained ["Connection","Open_Ok"] in
  send_Channel_Open ();
  let _ = recv_amqp_constrained ["Channel","Open_Ok"] in
  if interactive then begin
    send_Queue_Declare ~queue:prin ~durable:false();
    let _ = recv_amqp_constrained ["Queue","Declare_Ok"] in
    send_Basic_Consume ~queue:prin ~ctag:prin ();
    let _ = recv_amqp_constrained ["Basic","Consume_Ok"] in
    ()
  end;
  let listening () =
    try while !listening do
(*      Thread.yield();*)
      match recv_amqp_constrained 
        ["Basic","Deliver";"Queue","Declare_Ok";"Queue","Bind_Ok";
         "Exchange","Declare_Ok";"Basic","Consume_Ok"]
        with 
          |  (_,(("Queue","Declare_Ok"),_))
          |  (_,(("Queue","Bind_Ok"),_))
          |  (_,(("Exchange","Declare_Ok"),_))
          |  (_,(("Basic","Consume_Ok"),_)) -> ()
          |  (_,(("Basic","Deliver"),vparam)) -> begin
            let delivery_tag = match vparam with 
                [Pval_shortstr _ ;Pval_longlong i;Pval_bool _ ;Pval_shortstr _ ;Pval_shortstr _ ] -> i
              | _ -> assert false in
            debug_com (Printf.sprintf "Receiving Deliver message with tag %d" delivery_tag) ;
            let _ = read_amqp_message (recv_amqp ()) in
            let m = read_amqp_message (recv_amqp ()) in
            (match m with
                 CBody s -> Queue.push s recvqueue
               | _ -> assert false);
            send_Basic_Ack ~delivery_tag ()
          end
        | _ -> assert false
    done
    with _ -> ()
  in
  let t = Thread.create listening () in
  listening_thread := Some t;
  ()
    

let init queue =
  debug_com (Printf.sprintf "Initialising queue %s" queue) ;
  (*  List.iter (fun x -> *)
  send_Queue_Declare ~queue ~exclusive:false ~durable:false ();
  (*  let _ = recv_amqp_constrained ("Queue","Declare_Ok") in *)
  debug_com (Printf.sprintf "Done") ;
  ()
    (* ) queues *)
  
let init_session sessionid lr =
  debug_com (Printf.sprintf "Initialising exchange %s" sessionid) ;
  send_Exchange_Declare ~exchange:sessionid ~auto_delete:true ();
  List.iter (function (role,q) -> 
               send_Queue_Bind ~exchange:sessionid ~queue:q ~routing_key:role
                 ()) lr;
  ()
  

let close () =
  listening:=false;
  Thread.delay 1.0;
  (* ignore (match !listening_thread with
     None -> () | Some t -> Thread.kill t);*)
  send_Connection_Close ();
  let rec wait () =
    match recv_amqp_constrained
      ["Connection","Close_Ok";"Queue","Declare_Ok";
       "Queue","Bind_Ok";"Exchange","Declare_Ok";"Exchange","Delete_Ok"] with
        | (_,(("Connection","Close_Ok"),_)) -> ()
        | _ ->  (assert false ; Thread.delay 1.0;wait ())
  in
  debug_com (Printf.sprintf "Closing connection") ;
  (*wait (); *)
  Thread.delay 1.0;
  debug_com (Printf.sprintf "Connection closed") ;
  let s = descr_of_out_channel !out_amqp in
  debug_com (Printf.sprintf "Shutting down") ;
  Unix.shutdown s Unix.SHUTDOWN_ALL
  
  

let send ?(verb=true) (dest:string) (text:str): unit = 
  send_Basic_Publish ~exchange:"" ~routing:dest ~payload:text ();
  ()

let psend ?(verb=true) ?(exchange="") acktable mnb (dest:string) (text:str): unit = 
  send_Basic_Publish ~exchange:exchange ~routing:dest ~payload:text ();
  debug_com (Printf.sprintf "Message sent!" ) ;
  ()

let precv () : str option =
  if Queue.is_empty recvqueue
  then None
  else begin
    debug_com ("A message is received!") ;
    Some (Queue.pop recvqueue)
  end
  

let sendAck session srole sid rrole rid mnb =
  ()
