(********************************************************************)
(* SessML - Runtime                                                 *)
(*                                                                  *)
(* global.ml                                                        *)
(********************************************************************)
(* Time-stamp: <Pierre-Malo Denielou - 2010> *)

(**********************************)
(* Global constants and functions *)
(**********************************)

let debug_runtime = ref false
let debug_monitor = ref false
let debug_impl = ref true
let debug_com = ref false
let debug_com_full = ref false
let debug_com_io = ref false
let debug_ref = ref false
let debug_ack = ref false
let debug_concat = ref false

let delay_bool = ref false

let set_debug_runtime v = debug_runtime := v
let set_debug_monitor v = debug_monitor := v
let set_debug_impl v = debug_impl := v
let set_debug_com v = debug_com := v
let set_debug_com_full v = debug_com_full := v
let set_debug_com_io v = debug_com_io := v
let set_debug_ref v = debug_ref := v
let set_debug_ack v = debug_ack := v
let set_debug_concat v = debug_concat := v

let set_delay v = delay_bool := v

let newdebug debug_bool modul msg =
  if !debug_bool then begin
    print_string ("["^modul^"] ");
    print_string msg ;
    print_newline ();
    flush Pervasives.stdout
  end
  else ()

let make_debug = newdebug debug_impl
let debug_monitor = newdebug debug_monitor "Monitor"
let debug = newdebug debug_runtime "Runtime"
let debug_com = newdebug debug_com "Com"
let debug_com_full = newdebug debug_com_full "Com+"
let debug_com_io = newdebug debug_com_io "ComIO"
let debug_ack = newdebug debug_ack "Ack"
let debug_ref = newdebug debug_ref "Ref"
let debug_concat = newdebug debug_concat "Concat"


let delay = if !delay_bool then Thread.delay else ignore

(* Global transport option *)

let amqp_port = ref 5672 


let test_eq i j m =
  if i = j then ()
  else failwith m   

let test_inf i j m =
  if i < j then ()
  else failwith m   

(*********************)
(* List Manipulation *)
(*********************)

(* Substitution *)
let alias subst =
  function n ->
    if List.mem_assoc n subst 
    then List.assoc n subst
    else n

(* Remove duplicates in a list *)
let rec remove_duplicates l = 
  match  l with
      [] -> []
    | a::q -> if (List.mem a q) 
      then (remove_duplicates q) 
      else a::(remove_duplicates q)

(* Substraction *)
let rec substract l ll =
  let rec remove x = function
  [] -> []
    | a::q -> if a = x then remove x q else a::(remove x q) in
  match l with
      [] -> ll
    | a::q -> substract q (remove a ll)

let rec substract_eq l ll =
  let rec remove x = function
  [] -> []
    | a::q -> if a == x then remove x q else a::(remove x q) in
  match l with
      [] -> ll
    | a::q -> substract q (remove a ll)


(* Split on newlines *)
let split s =
  let rec aux i =
    if i = String.length s then
     []
    else
      try 
        let n = String.index_from s i '\n' in
        let s = String.sub s i (n-i) in
        if s="" then aux (n+1) else s::(aux (n+1))
      with Not_found -> 
        let s = String.sub s i (String.length s -i) in
        if s = "" then [] else [s]
  in aux 0


(**********)
(* Base64 *)
(**********)

let char_table = 
  [|'A';'B';'C';'D';'E';'F';'G';'H';'I';'J';'K';'L';'M';'N';'O';'P';'Q';'R';'S';
    'T';'U';'V';'W';'X';'Y';'Z';'a';'b';'c';'d';'e';'f';'g';'h';'i';'j';'k';'l';
    'm';'n';'o';'p';'q';'r';'s';'t';'u';'v';'w';'x';'y';'z';'0';'1';'2';'3';'4';
    '5';'6';'7';'8';'9';'+';'/'|] 

let table_back c =
  match c with
    | 'a' .. 'z' -> Some (int_of_char c - 71)
    | 'A' .. 'Z' -> Some (int_of_char c - 65)
    | '0' .. '9' -> Some (int_of_char c + 4)
    | '+' -> Some (62)
    | '/' -> Some (63)
    | '=' -> None
    | _ -> assert false

let eq = '='
  
let base64 (s:string) =
  let size = 8*String.length s in
  let ssize = 
    (size / 24 ) * 4 + (if size mod 24 <> 0 then 4 else 0) in
  let res = String.create ssize in
  (*  let _ = Printf.printf "Taille : %i donc %i\n" size ssize in *)
  let rec encode p =
    if p*8 >= size then ()
    else
      let byte1 = int_of_char (s.[p]) in
      let bits1 = (byte1 / 4) in
      let _ = res.[p*8/6] <- (char_table.(bits1)) in
      if 8*p+8 >= size 
      then 
        let bits2 = (byte1 mod 4)*16 in
        let _ = res.[p*8/6+1] <- (char_table.(bits2)) in
        ()
      else
        let byte2 = int_of_char (s.[p+1]) in
        let bits2 = (byte1 mod 4)*16 + (byte2 / 16) in
        let _ = res.[p*8/6+1] <- (char_table.(bits2)) in
        if 8*p+16 >= size 
        then
          let bits3 = (byte2 mod 16)*4 in
          let _ = res.[p*8/6+2] <- (char_table.(bits3)) in
          ()
        else 
          let byte3 = int_of_char (s.[p+2]) in
          let bits3 = (byte2 mod 16)*4 + (byte3 / 64) in
          let bits4 = (byte3 mod 64) in
          let _ = res.[p*8/6+2] <- (char_table.(bits3)) in
          let _ = res.[p*8/6+3] <- (char_table.(bits4)) in
          encode (p+3)
  in
  (* filling the last bits with equals *)
  if size mod 24 <> 0 then 
    for i = ssize - 3 to ssize-1 do
      res.[i] <- eq
    done;
  encode 0 ;
  res

    
let ibase64 (s:string) =
  let size = String.length s in
  let ssize = size * 6 / 8 in
  let res = String.create ssize in
  
  let rec decode p =
    if p = size 
    then 0
    else
      let i = (p*6)/8 in
      let v1 = table_back s.[p] in
      let v2 = table_back s.[p+1] in
      let v3 = table_back s.[p+2] in
      let v4 = table_back s.[p+3] in
      match (v1,v2,v3,v4) with
        | Some a, Some b, Some c, Some d -> 
            (*            let _ = Printf.printf "Chars : %i %i %i %i\n" a b c d in
                          let _ = Printf.printf "donne : %i %i %i \n" (a * 4 + b / 16) ((b
                          mod 16)*16 + c / 4) ((c mod 4)*64 + d) in *)
          let _ = res.[i] <- char_of_int (a * 4 + b / 16) in
          let _ = res.[i+1] <- char_of_int ((b mod 16)*16 + c / 4) in
          let _ = res.[i+2] <- char_of_int ((c mod 4)*64 + d) in
          decode (p+4)
        | Some a, Some b, Some c, None ->
          let _ = res.[i] <- char_of_int (a * 4 + b / 16) in
          let _ = res.[i+1] <- char_of_int ((b mod 16)*16 + c / 4) in
          1
        | Some a, Some b, None, None ->
          let _ = res.[i] <- char_of_int (a * 4 + b / 16) in
            (*            let _ = res.[i+1] <- char_of_int (b mod 16) in *)
          2 
        | _ -> assert false
  in
  let i = decode 0 in
  String.sub res 0 (ssize-i)

(************)
(* My Queue *)
(************)

module Myq =
  (struct
     type 'a t = 'a list ref

     let create () = 
       let q = ref [] in
       q

     let push q x = q:= List.append !q [x]

     let pop q = match !q with
         [] -> assert false
       | a::r -> q:= r; a

     let peek q = 
       List.hd (!q)

     let move q f =
       let rec revapp l r = match l with
           [] -> r
         | a::q -> revapp q (a::r) in
       let rec aux l = function 
           [] -> None
         | a::r -> 
             if f a 
             then Some (a::(revapp l r))
             else aux (a::l) r
       in
       match aux [] !q with
           None -> false
         | Some l -> (q:= l ; true)

     let is_empty q = (!q = [])

     let length q = List.length !q
   end:
    sig
      type 'a t
      val create : unit -> 'a t
      val push : 'a t -> 'a -> unit
      val pop : 'a t -> 'a
      val peek : 'a t -> 'a
      val move : 'a t -> ('a -> bool) -> bool
      val is_empty: 'a t -> bool
      val length : 'a t -> int
    end 
  )

