(********************************************************************)
(* SessML - Runtime                                                 *)
(*                                                                  *)
(* runtime.ml                                                       *)
(********************************************************************)
(* Time-stamp: <Deniélou Malo - 2011> *)

open Unix
open Global
open Data
open Prins

module Transport = Amqp

open Transport

let set_amqp_port port = 
  amqp_port := port 

let mnb_check = ref false

let set_mnb_check b =
  mnb_check := b

(****************)
(* Runtime loop *)
(****************)

let runtime = ref None

type sendmsg = 
  | Wired of string*string*string list*string*string*string*string*string
      (* (session,kind,lt,srole,sid,rrole,rid,content) *)
  | Simple of string*string*string list*string*string*string*string*string

let sendqueue : sendmsg Queue.t = Queue.create ()
let cache = Hashtbl.create 10
let sessiontable = Hashtbl.create 10
let recvqueue = ref []
let acktable = Hashtbl.create 10

let runtimeloop () = 
  debug ("Starting runtime loop") ;
  while true do
    Thread.yield ();
    if Hashtbl.length acktable>0 || Queue.is_empty sendqueue 
    then 
      (begin (* Receive *)
         let m = precv () in 
         match m with 
           | Some m -> 
               (begin
                  let msg = readmsg m in
                  (match msg.lbl with
                     | "ack" -> (
                         debug_ack (sprintf_msg msg);
                         if Hashtbl.mem acktable msg.mnb 
                         then (debug_ack ("Ack "^msg.mnb^" received") ;
                               Hashtbl.remove acktable msg.mnb)
                         else (debug_ack ("Unknown ack received ... ignoring");
                               (*assert false *) ))
                     | "inv" -> (
                         debug (sprintf_msg msg);
                         if !mnb_check && Hashtbl.mem cache msg.mnb
                         then (debug_ack 
                                 (Printf.sprintf " already received message %s: acking again to %s"
                                    msg.mnb msg.sid);
                               ignore (sendAck msg.session msg.rrole msg.rid msg.srole msg.sid msg.mnb))
                         else (
                           ignore (sendAck msg.session msg.rrole msg.rid msg.srole msg.sid msg.mnb);
                           Hashtbl.add cache msg.mnb ();
                           let sessionid = msg.session in
                           let role = msg.rrole in
                           let session = msg.body in
                           if Hashtbl.mem sessiontable (sessionid,(role,session))
                           then (
                             debug ("already received invitation") ;
                             assert false
                           ) else (
                             debug ("Receiving invitation") ;
                             Hashtbl.add sessiontable (sessionid,(role,session)) sessionid; 
                             let lref = 
                               if List.mem_assoc sessionid !recvqueue 
                               then List.assoc sessionid !recvqueue 
                               else (let l = ref [] in 
                                     recvqueue := (sessionid,l)::!recvqueue ;
                                     l) in
                             let () = 
                               if List.mem_assoc role !lref
                               then failwith "Role already locally running"
                               else (let q = Myq.create () in
                                     lref := (role,q)::!lref) in
                             debug (Printf.sprintf 
                                      "Pushing invitation to session queue of %s" role) ;
                             let squeue = List.assoc role !(List.assoc session !recvqueue) in
                             Myq.push squeue (msg.srole,msg.sid,msg.rrole,msg.rid,msg.lbl,msg.lv,sessionid)
                           )))
                     | "reg"-> (
                         debug (sprintf_msg msg);
                         if !mnb_check && Hashtbl.mem cache msg.mnb
                         then (
                           ignore (sendAck msg.session msg.rrole msg.rid msg.srole msg.sid msg.mnb ))
                         else ( (* If not already received *)
                           ignore (sendAck msg.session msg.rrole msg.rid msg.srole msg.sid msg.mnb);
                           Hashtbl.add cache msg.mnb ();
                           let content,greg = iconcat msg.body in
                           reg_read greg;
                           let session,sreg = iconcat content in
                           let session,sessionid = iconcat session in
                           let () = 
                             if List.mem_assoc msg.session !recvqueue 
                             then (let lref = List.assoc msg.session !recvqueue in
                                   let q = List.assoc msg.rrole !lref in 
                                   (* forgot the case where there are two roles played
                                      by the same principal in the same session *)
                                   debug ("Putting reg message in the corresponding queue") ;
                                   (* If not already received *)
                                   (*            sendAck session srole sid rrole rid mnb;*)
                                   Myq.push q (msg.srole,msg.sid,msg.rrole,msg.rid,msg.lbl,msg.lv,content)
                                  ) else (
                               let l = ref [] in 
                               recvqueue := (msg.session,l)::!recvqueue;
                               let q = Myq.create () in
                               l := (msg.rrole,q)::!l ;
                               let lref = List.assoc session !recvqueue in
                               let q = List.assoc msg.rrole !lref in
                               debug ("Putting reg message in the corresponding queue") ;
                               (* If not already received *)
                               (*            sendAck session srole sid rrole rid mnb;*)
                               Myq.push q (msg.srole,msg.sid,msg.rrole,msg.rid,msg.lbl,msg.lv,content))
                           in ()))
                     | _ -> (
                         debug (sprintf_msg msg);
                         if !mnb_check && Hashtbl.mem cache msg.mnb
                         then (debug_ack 
                                 (Printf.sprintf " already received message %s: acking again to %s"
                                    msg.mnb msg.sid);
                               ignore (sendAck msg.session msg.rrole msg.rid msg.srole msg.sid msg.mnb))
                         else ( (* If not already received *)
                           ignore (sendAck msg.session msg.rrole msg.rid msg.srole msg.sid msg.mnb);
                           Hashtbl.add cache msg.mnb ();
                           let lref = List.assoc msg.session !recvqueue in
                           let q = List.assoc msg.rrole !lref in
                           Myq.push q (msg.srole,msg.sid,msg.rrole,msg.rid,msg.lbl,msg.lv,msg.body)
                         ))
                         (*   | _ -> assert false *)
                  )
                end)
           | None -> (
               if Hashtbl.length acktable>0 
               then
                 (begin
                    debug_ack ("Resending previous messages") ;
                    Hashtbl.iter 
                      (fun mnb -> function (rid,txt) -> 
                         debug_ack (Printf.sprintf "Resending to %s message %s"
                                      rid mnb);
                         send ~verb:false rid txt)
                      acktable ;
                    debug_ack ("Done")
                  end)
               else  (Thread.yield ()) (* (print_string ".")  *)  )
       end)
    else (* 
            if not (Queue.is_empty sendqueue || Hashtbl.length acktable>0) then *)
      begin (* Send *)
        match Queue.pop sendqueue with
            Wired (session,kind,lt,srole,sid,rrole,rid,content) -> 
              let mnb = string_of_int (Random.bits ()) in
              let txt = buildmsg session mnb srole sid rrole rid kind lt content in
              (* let () = Hashtbl.add acktable mnb (rid,txt) in *)
              debug (Printf.sprintf "Sending message %s from %s:%s to %s:%s" 
                       mnb srole sid rrole rid) ;
              psend acktable mnb rid txt
          | Simple (session,kind,lt,srole,sid,rrole,rid,content) -> 
              let mnb = string_of_int (Random.bits ()) in
              let txt = buildmsg session mnb srole sid rrole rid kind lt content in
              debug (Printf.sprintf "Sending message %s from %s:%s to %s:%s" 
                       mnb srole sid rrole rid) ;
              psend ~exchange:session acktable mnb rrole txt
      end;
    delay 1. ;
    ()
  done

let local_id = ref []

let start_runtime ?(interactive=true) session role id =
  let lref = 
    if List.mem_assoc session !recvqueue
    then List.assoc session !recvqueue 
    else (let l = ref [] in recvqueue := (session,l)::!recvqueue ; l)
  in 
  let () = if List.mem_assoc role !lref
    then failwith "Role already locally awaited"
    else (lref := (role,Myq.create ())::!lref)
  in
  let () = if List.mem id !local_id
    then () 
    else (
      local_id := id:: !local_id;
      ignore(bind ~interactive id)) in
  if !runtime = None 
  then (
    Random.self_init ();
    runtime := Some (Thread.create runtimeloop ())
  )
  else ()

let close_runtime () =
  while ((not (Queue.is_empty sendqueue)) || Hashtbl.length acktable>0) do
    Thread.yield ();
    ()
  done;
  debug ("Ack to quit message received: closing ...") ;
  close ()

(*******)
(* API *)
(*******)

let sendMsg session srole sid rrole rid tag lt v sub =
  let lt = List.map (alias sub) lt in
  Queue.push (Wired (session,tag,lt,srole,sid,rrole,rid,v)) sendqueue

let sendSMsg session srole sid rrole rid tag lt v sub =
  let lt = List.map (alias sub) lt in
  Queue.push (Simple (session,tag,lt,srole,sid,rrole,rid,v)) sendqueue

let recvMsg session rrole tags =
  debug (Printf.sprintf "Waiting for reception for role %s in session %S" rrole session) ;
  let lref = List.assoc session !recvqueue in
  let q = List.assoc rrole !lref in
  let wait = ref true in
  while !wait do
    Thread.yield ();
    while Myq.is_empty q do (delay 1. ; Thread.yield ()) done;
    debug ("Tentative reception") ;
    let there = (Myq.move q 
                   (function (srole,sid,rrole,rid,tag,lv,body) ->
                      List.mem (tag,lv) tags)) in
    if there then (
      debug ("Message on top of the queue") ; 
      wait := false)
    else (delay 1.);
  done;
  let (srole,sid,rrole,rid,tag,lv,body) = Myq.pop q in
  debug_com (Printf.sprintf "Reading received message: tag is %s, and body is %s" tag body) ;
  ((tag,lv),body)


let recvInv session rrole =
  debug (Printf.sprintf "Expecting invitation for role %s" rrole) ;
  let (_,sessionid) = recvMsg session rrole ["inv",[]] in
  sessionid
  
  
let sendInv session sessionid sid rrole rid =
  debug ("Sending invitation message") ;
  let body = concat session rrole in
  Queue.push (Wired (sessionid,"inv",[],"Adm",sid,rrole,rid,body)) sendqueue
  
let sendSInv session sessionid sid reg =
  debug ("Sending invitation message") ;
  let lr = List.map (fun (r,p) -> (r,sessionid^r)) reg in
  List.iter (fun (r,q) -> init q) lr;
  init_session sessionid lr;
  List.iter (fun (r,p) ->
              (* init (List.hd p);*)
               let body = session in
               Queue.push (Wired(sessionid,"inv",[],"Adm",sid,r,List.hd p,body))
                 sendqueue) reg

let sendAckJoin session srole sid regid =  
  (* Should bind the session:srole *)
  debug ("Consuming from session queue") ;
  send_Basic_Consume ~queue:(session^srole) ~ctag:(session^srole) ();
  ()

let sendJoin session srole sid regid = 
  debug ("Queuing join message") ;
  let ip,po = List.assoc sid !addrDb in
  let content = concat sid (concat ip (string_of_int po)) in
  Queue.push (Wired (session,"join",[],srole,sid,"Adm",regid,content)) sendqueue;
  debug ("Expecting registry update") ;
  let (_,content) = recvMsg session srole ["reg",[]] in
  let session,reg = iconcat content in
  let session,sessionid = iconcat session in
  debug ("Received Registry update") ;
  debug (Printf.sprintf "reading list : %s" reg) ;
  let ls = iconcatlist reg in
  debug (Printf.sprintf "reading list of %d elements" (List.length ls)) ;
  let l = List.map 
    (fun x -> let role,pl = iconcat x in (role,iconcatlist pl)) ls in
  debug ("Received Registry update completed") ;
  sessionid,l

let sendQuit session srole sid regid = 
  debug ("Queuing quit message") ;
  let content = sid in
  Queue.push (Wired (session,"quit",[],srole,sid,"Adm",regid,content)) sendqueue
  (* delay(1.0)  To be sure that the messages are through *)
 
let sendLock session srole sid regid = 
  debug ("Queuing lock message") ;
  let content = sid in
  Queue.push (Wired (session,"lock",[],srole,sid,"Adm",regid,content)) sendqueue;
  debug ("Expecting registry update") ;
  let (_,content) = recvMsg session srole ["reg",[]] in
  debug ("Received Registry update") ;
  let _,reg = iconcat content in
  debug (Printf.sprintf "reading list : %s" reg) ;
  let ls = iconcatlist reg in
  debug (Printf.sprintf "reading list of %d elements" (List.length ls)) ;
  let l = List.map 
    (fun x -> let role,pl = iconcat x in (role,iconcatlist pl)) ls in
  debug ("Received Registry update completed") ;
  l

let sendUnlock session srole sid regid  = 
  debug ("Queuing unlock message") ;
  let content = sid in
  Queue.push (Wired (session,"unlock",[],srole,sid,"Adm",regid,content)) sendqueue

let sendReg session sessionid sid rrole rid reg =
  debug ("Sending registry message") ;
  let lr = List.map (function r,lp -> concat r (concatlist lp)) reg in
  let sreg = concatlist lr in
  let greg = reg_write reg in
  let body = concat (concat (concat session sessionid) sreg) greg in
  Queue.push (Wired (sessionid,"reg",[],"Adm",sid,rrole,rid,body)) sendqueue
  
type admin = 
  | AdmJoin of string * string
  | AdmQuit of string * string
  | AdmLock of string * string
  | AdmUnlock of string * string

let recvAdm session sessionid =
  debug_com ("Waiting for Adm reception") ;
  let lref = List.assoc session !recvqueue in
  let q = List.assoc "Adm" !lref in
  let lref = List.assoc sessionid !recvqueue in
  let qq = List.assoc "Adm" !lref in
  debug_com (Printf.sprintf "Looking in Adm queue for session %S and %S"
               sessionid session) ;
  while Myq.is_empty q && Myq.is_empty qq do
    (Thread.yield ();
     delay 1. )
     (*  debug_com ("Still waiting for a message")) *)
  done;
  delay 1. ; 
  let (srole,sid,rrole,rid,tag,_,body) = 
    if Myq.is_empty qq then Myq.pop q else Myq.pop qq in
  debug_com (Printf.sprintf "An Adm message %s in the queue" tag) ;
  match tag with
    | "join" -> AdmJoin(srole,sid)
    | "quit" -> AdmQuit(srole,sid)
    | "lock" -> AdmLock(srole,sid)
    | "unlock" -> AdmUnlock(srole,sid)
    | _ -> assert false 


let update_reg_reglock locked locking reglock jq reg session sessionid id condition roles =
  let jjq = Queue.create () in
  Queue.iter 
    (function
       | ("Quit",(r,p)) -> (
           debug (Printf.sprintf "%s leaves the session as %s" p r);
           let l = List.assoc r !reg in
           let rreg = List.remove_assoc r !reg in
           reg:= (r,List.filter (fun x->x<>p) l)::rreg)
       | _ -> ())
    jq;
  Queue.iter 
    (function
       | ("Join",(r,p)) -> (
           debug (Printf.sprintf "%s wants to join as %s" p r);
           if List.mem_assoc r !reg
           then (let l = List.assoc r !reg in
                 let rreg = List.remove_assoc r !reg in
                 if (condition r (1+ List.length l) >= 0)
                 then (debug (Printf.sprintf "%s joins the session as %s" p r);
                       reg:= (r,p::l)::rreg;
                       reglock:=(r,p)::!reglock
                      )
                 else (debug (Printf.sprintf "%s cannot join as %s!" p r);
                       Queue.push ("Join",(r,p)) jjq))
           else (debug (Printf.sprintf "%s joins the session as %s" p r);
                 reg:= (r,[p])::!reg;
                 reglock:=(r,p)::!reglock
                ))
       | ("Quit",(r,p)) -> ()
       | _ -> assert false)
    jq;
  Queue.clear jq;
  Queue.iter (fun x -> Queue.push x jq) jjq;
  if (List.for_all 
        (fun role -> condition role (List.length (List.assoc role !reg)) = 1) roles) &&
    (List.for_all (fun (r,l) -> List.for_all (fun y -> List.mem (r,y) !reglock) l) !reg)
  then (locked:=true ; locking:=false;
        debug_ref (Printf.sprintf "locked = true; locking = false");
        debug (Printf.sprintf "~~~ Session in a locked state ~~~ ");
        List.iter (function (role,rid) -> sendReg session sessionid id role rid !reg)
          !reglock)
  else (debug (Printf.sprintf "Session not yet locked"))



let lock locked locking reglock jq reg session sessionid id role rid condition roles =
  if (not !locked)
  then begin
    debug (Printf.sprintf "%s asks to lock the session as %s" rid role); 
    if not !locking
    then (locking:=true;locked:=false;
          debug_ref (Printf.sprintf "locked = false; locking = true");
          debug (Printf.sprintf "~~~ Starting a new iteration ~~~"));
    debug_ref (Printf.sprintf "Updating reglock");
    if List.mem_assoc role !reg && List.mem rid (List.assoc role !reg)
    then (debug (Printf.sprintf "Registering a lock from %s:%s" role rid);
          reglock:=(role,rid)::!reglock);
    update_reg_reglock locked locking reglock jq reg session sessionid id condition roles;
    debug (Printf.sprintf "Session in a locking state");
    ()
  end
  else
    (
      let lref = List.assoc session !recvqueue in
      let q = List.assoc "Adm" !lref in
      Myq.push q (role,rid,"Adm",id,"lock",[],rid)
    )


let unlock locked locking reglock jq reg session sessionid id role rid = 
  if (!locked && not !locking)
  then begin
    debug (Printf.sprintf "%s asks to unlock the session as %s" rid role);
    let newreg = List.filter (fun x -> x <> (role,rid)) !reglock in
    debug_ref (Printf.sprintf "Updating reglock");
    reglock:= newreg;
    if !reglock = []
    then (locked:=false;locking:=false;
          debug_ref (Printf.sprintf "locked = false");
          debug_ref (Printf.sprintf "locking = false");
          debug (Printf.sprintf "~~~ Session is not locked anymore ~~~"))
  end
  else begin
    let lref = List.assoc session !recvqueue in
    let q = List.assoc "Adm" !lref in
    Myq.push q (role,rid,"Adm",id,"unlock",[],rid)
  end
  

let join locked locking reglock jq reg session sessionid id (role,rid) condition roles =
  debug (Printf.sprintf "%s asks to join the session as %s" rid role);
  Queue.push ("Join",(role,rid)) jq;
  lock locked locking reglock jq reg session sessionid id role rid condition roles

let quit locked locking reglock jq reg session sessionid id (role,rid) condition roles =
  Queue.push ("Quit",(role,rid)) jq;
  if (not !locked)
  then begin
    debug (Printf.sprintf "%s asks to quit the session as %s" rid role);
    update_reg_reglock locked locking reglock jq reg session sessionid id condition roles
  end
  else
    begin
(*    let lref = List.assoc session !recvqueue in
    let q = List.assoc "Adm" !lref in
    Myq.push q (role,rid,"Adm",id,concat "Quit" rid)
*)
      ()
  end
  
  
