(********************************************************************)
(* SessML - Monitor                                                 *)
(*                                                                  *)
(* monitor.ml                                                       *)
(********************************************************************)
(* Time-stamp: <Deniélou Malo - 2011> *)


open Global
open Data
open Syntax

let () = set_debug_monitor true

let debug = debug_monitor

let _ = set_debug_monitor true
let _ = set_debug_com true
let _ = set_debug_com_io false

(********************************)
(* Monitor states  & projection *)
(********************************)


type tempmsg = 
    Client of (string*string*string*session_message) 
  | Server of (string*string*string*session_message)

type state = {mutable lt:local_type ; 
              lq : (string * (tempmsg Queue.t)) list;
              nt : (lnode,local_type option) Hashtbl.t}

let invitationtable = Hashtbl.create 10 
let creationtable = Hashtbl.create 10 
let roletable : (string*string,state) Hashtbl.t = Hashtbl.create 10

let monitor_lock = Mutex.create ()

let notification_lock = Mutex.create ()
let notification_condition = Condition.create ()
let notification = Queue.create ()

type monitor = Mon of (tempmsg -> bool * monitor) | Moff


let read session =
  let lexbuf = Lexing.from_string session in
  debug "Lexer built" ;
  let sessionast = Parser.sess Lexer.token lexbuf in
  match sessionast with
    | Globalast (session_name,rl,g) -> begin
        let g = Wf.sanitize rl g in
        rl,g
      end
    | _ -> assert false
  

let project g role =
  let proj = Wf.torole role g in
  let proj = Wf.clean_local_role proj in 
  proj

let build_node_table local_type =
  let node_table = Hashtbl.create 10 in
  let rec build_table = function
    | LEnd n ->  (Hashtbl.add node_table n None)
    | LStop n -> (Hashtbl.add node_table n None)
    | LGoto n -> ()
    | LSend (n,((r,rr),ml)) -> (
        Hashtbl.add node_table n (Some (LSend (n,((r,rr),ml))));
        List.iter (function (l,lv,lt) -> build_table lt) ml
      )
    | LRecv (n,((s,ss),ml)) -> (
        Hashtbl.add node_table n (Some (LRecv (n,((s,ss),ml))));
        List.iter (function (l,lv,lt) -> build_table lt) ml
      )
    | LPoll (n,((x,r),pl,g,c)) ->  ( assert false  )
    | LPar (n,(g,r,c)) -> (
        Hashtbl.add node_table n (Some (LPar (n,(g,r,c))));
        build_table g; build_table r; build_table c
      )
    | LRec (n,g) -> (
        Hashtbl.add node_table n (Some (LRec (n,g)));
        build_table g
      )
    | LMu (n,g) ->  (
        Hashtbl.add node_table n (Some (LMu (n,g)));
        build_table g
      )
    | LStart (n,g) -> (
        Hashtbl.add node_table n (Some (LStart (n,g)));
        build_table g
      )
    | LInv (n,g) -> (
        Hashtbl.add node_table n (Some (LInv (n,g)));
        build_table g
      )
  in
  node_table


(*****************)
(* Communication *)
(*****************)

let client_out = ref stdout
let client_in = ref Unix.stdin
let client_messages = Queue.create ()

let send_client m =
  output_string !client_out m;
  flush !client_out;
  debug_com (Printf.sprintf "Message sent to client!" ) ;
  ()

let send_amqp = Amqp.send_amqp

let expecting_ack = ref false

let forward_client (mdeliver,mcheader,mcontent) =
  debug (Printf.sprintf "Forwarding a correct message to the client");
  send_client mdeliver;
  send_client mcheader;
  send_client mcontent;
  debug (Printf.sprintf "Forwarding done");
  expecting_ack := true
(*  ignore (Amqp.recv_amqp_constrained 
            ~in_amqp:client_in ~messages:client_messages
    ("Basic","Ack")) *)

let forward_server (mdeliver,mcheader,mcontent) =
  debug (Printf.sprintf "Forwarding a correct message to the server");
  send_amqp mdeliver;
  send_amqp mcheader;
  send_amqp mcontent;
  debug (Printf.sprintf "Forwarding done");
(*  ignore (Amqp.recv_amqp_constrained ["Basic","Ack"])*)
  ()

let listening_server = ref true 
let listening_client = ref true 
  
let listening_server_thread = ref None 
let listening_client_thread = ref None 
  
let client_temp = Queue.create () 
let server_temp = Queue.create () 
let client_lock = Mutex.create ()
let client_condition = Condition.create ()
let server_lock = Mutex.create ()
let server_condition = Condition.create ()

let treating_server = ref true 
let treating_client = ref true 

let treating_server_thread = ref None 
let treating_client_thread = ref None 

let monitor_thread = ref None

let monitoring = ref true

let listening_server () =
  try
  while !listening_server do
    match (Amqp.recv_amqp_constrained 
             ~stop:(fun () -> not !listening_server)
      ["Basic","Deliver";"Connection","Close_Ok";"Queue","Delete_Ok";"Basic","Consume_Ok";
       "Queue","Declare_Ok";"Queue","Bind_Ok";"Exchange","Declare_Ok"]) with
  (*          None -> (Thread.yield ();delay 0.1) *)
        | (mdeliver,(("Basic","Deliver"),vparam)) -> 
            begin
              debug (Printf.sprintf "Receiving a message from the server");
              let delivery_tag = match vparam with 
                  [Amqp.Pval_shortstr _ ;
                   Amqp.Pval_longlong i;
                   Amqp.Pval_bool _ ;
                   Amqp.Pval_shortstr _ ;
                   Amqp.Pval_shortstr _ ] -> i
                | _ -> assert false in
              let mcheader = Amqp.recv_amqp () in
              let mcontent = Amqp.recv_amqp () in
              (match Amqp.read_amqp_message mcontent with
                  Amqp.CBody s -> ( 
                    Mutex.lock server_lock;
                    Queue.push (mdeliver,mcheader,mcontent,s) server_temp;
                    Mutex.unlock server_lock;
                    Condition.signal server_condition)
                 | _ -> assert false);
              Amqp.send_Basic_Ack ~delivery_tag ();
              ()
            end
        |  (mcloseok,(("Connection","Close_Ok"),vparam)) ->
            begin
              send_client mcloseok;
              listening_server:= false;
              listening_client:= false;
              treating_server:= false;
              treating_client:= false;
              monitoring:=false;
              Condition.signal client_condition;
              Condition.signal server_condition;
              Tcp.listen:=false
            end
        |  (mconsumeok,(("Basic","Consume_Ok"),vparam)) ->
            debug (Printf.sprintf "Receiving a consumeok message from the server");
            send_client mconsumeok
        |  (m,(("Queue","Delete_Ok"),vparam))
        |  (m,(("Queue","Declare_Ok"),vparam))
        |  (m,(("Queue","Bind_Ok"),vparam))
        |  (m,(("Exchange","Declare_Ok"),vparam)) -> 
            begin
              debug (Printf.sprintf "Receiving an administrative message from the server");
              send_client m
            end
        | _ -> assert false
  done
  with Amqp.Connection_stop -> (debug ("Closing the server listening thread ...") ;)

let listening_client () =
  try
  while !listening_client do
    match Amqp.recv_amqp_constrained 
      ~in_amqp:client_in ~messages:client_messages
      ~stop:(fun () -> not !listening_client)
      ["Basic","Publish";"Connection","Close";"Queue","Delete";"Basic","Ack"; 
       "Basic","Consume";"Queue","Declare";"Queue","Bind";"Exchange","Declare"]
    with 
  (*          None -> (Thread.yield ();delay 0.1) *)
        | (mdeliver,(("Basic","Publish"),vparam)) -> 
          begin
            debug (Printf.sprintf "Receiving a message from the client");
            let mcheader = Amqp.recv_amqp 
              ~in_amqp:client_in
              ~messages:client_messages () in
            let mcontent = Amqp.recv_amqp 
              ~in_amqp:client_in
              ~messages:client_messages () in
            (match Amqp.read_amqp_message mcontent with
                Amqp.CBody s -> ( 
                    Mutex.lock client_lock;
                    Queue.push (mdeliver,mcheader,mcontent,s) client_temp;
                    Mutex.unlock client_lock;
                    Condition.signal client_condition)
              | _ -> assert false);
            ()
          end 
        |  (mconsume,(("Basic","Consume"),vparam)) -> 
          begin
            debug ("Receiving join message (as a consume) from client") ;
            match vparam with
                [_;Amqp.Pval_shortstr queuename;Amqp.Pval_shortstr ctag;_;_] -> (
                  let session = String.sub queuename 0 4 in
                  let role = String.sub queuename 4 (String.length queuename -4)
                  in
                  let rl,g = Hashtbl.find invitationtable (session,role) in
                  let localtype = project g role in
                  let node_table = build_node_table localtype in
                  let lq = ("Adm",Queue.create ())::
                    (List.map (fun (r,_,_)-> (r,Queue.create())) rl) in
                  Hashtbl.add roletable (session,role) 
                    {lt=localtype;nt=node_table;lq=lq};
                  send_amqp mconsume
                )
              | _ -> assert false
          end 
        |  (mclose,(("Connection","Close"),vparam)) ->
          begin
            debug (Printf.sprintf "Receiving a closing message from the client");
            send_amqp mclose
          end
(*        | Some (mack,(("Basic","Ack"),vparam)) -> 
          if !expecting_ack then expecting_ack := false else assert false*)
        |  (m,(("Basic","Ack"),vparam)) -> ()
        |  (m,(("Queue","Declare"),vparam))
        |  (m,(("Queue","Delete"),vparam))
        |  (m,(("Queue","Bind"),vparam))
        |  (m,(("Exchange","Declare"),vparam)) -> 
          begin
            debug (Printf.sprintf "Receiving an administrative message from the client");
            send_amqp m
          end
        | _ -> assert false
  done
  with Amqp.Connection_stop -> (debug ("Closing the client listening thread ...") ;)

(**************)
(* Monitoring *)
(**************)


let treat_server () =
  try
    while !treating_server do
      Mutex.lock server_lock;
      if Queue.is_empty server_temp then
        Condition.wait server_condition server_lock;
      debug (Printf.sprintf "Treating a message from the server");
      let (mdeliver,mcheader,mcontent,m) = Queue.pop server_temp in
      Mutex.unlock server_lock;
      (*    if Queue.is_empty server_temp
            then (Thread.yield ();delay 0.4)
            else *)
      begin
        Mutex.lock monitor_lock;
        let msg = readmsg m in
        (match msg.lbl with
           | "ack" -> () (* No acks with AMQP *)
           | "inv" -> 
               begin
                 debug (sprintf_msg msg);
                 debug ("Receiving invitation from server") ;
                 let sessionid = msg.session in
                 let role = msg.rrole in
                 let session = msg.body in
                 if Hashtbl.mem invitationtable (sessionid,role)
                 then (
                   debug ("already received invitation, dropping ...")
                 )
                 else (
                   debug ("accepting invitation ...");
                   Hashtbl.add invitationtable (sessionid,role) (read session);
                   forward_client (mdeliver,mcheader,mcontent)
                 )
               end
           | _ -> (
               debug (Printf.sprintf "Putting %s in the queue of session %s"
                        (sprintf_msg msg) msg.session);
               let state = Hashtbl.find roletable (msg.session,msg.rrole) in
               let q = List.assoc msg.srole state.lq in
               Queue.push (Server (mdeliver,mcheader,mcontent,msg)) q;
               Mutex.lock notification_lock;
               Queue.push (msg.session,msg.rrole,q) notification;
               Mutex.unlock notification_lock;
               Condition.signal notification_condition
             ) (* Put in the right queue + notification queue *)
        );
        Mutex.unlock monitor_lock
      end
    done
  with Queue.Empty -> (debug ("Closing the server thread ...") ;)
 
let treat_client () =
  try
    while !treating_client do
      Mutex.lock client_lock;
      Condition.wait client_condition client_lock;
      debug (Printf.sprintf "Treating a message from the client");
      let (mdeliver,mcheader,mcontent,m) = Queue.pop client_temp in
    debug (Printf.sprintf "There are %d messages left n client_temp" (Queue.length client_temp));
      Mutex.unlock client_lock;
      (*    if Queue.is_empty client_temp
            then (Thread.yield ();delay 0.4)
            else *)
      begin
        Mutex.lock monitor_lock;
        (*     debug (Printf.sprintf "Treating a message from the client");
               let (mdeliver,mcheader,mcontent,m) = Queue.pop client_temp in*)
        let mytreat (mdeliver,mcheader,mcontent,m) =
          let msg = readmsg m in
          (match msg.lbl with
            | "ack" -> () (* No acks with AMQP *)
            | "inv" -> 
              begin
                debug (sprintf_msg msg);
                debug ("Receiving invitation from client") ;
                let sessionid = msg.session in
                let role = msg.rrole in
                let session = msg.body in
                if Hashtbl.mem invitationtable (sessionid,role)
                then (
                  debug ("sending already received invitation, forgetting ...");
                  Hashtbl.remove invitationtable (sessionid,role);
                  forward_server (mdeliver,mcheader,mcontent)
                )
                else (
                  if Hashtbl.mem creationtable (sessionid,session)
                  then (
                    debug ("Invitation sent"); 
                  (* I do not check if the invitation matches the role ... *)
                  ) else (
                    debug ("Creating session for invitation");
                    Hashtbl.add creationtable  (sessionid,session)
                      sessionid;
                    let lq = ["Adm",Queue.create ()] in
                    Hashtbl.add roletable (sessionid,"Adm")  
                      {lt=LStop (0,0);nt=Hashtbl.create 1;lq=lq}
                  );
                  debug ("Invitation sent"); 
                  forward_server (mdeliver,mcheader,mcontent)
                )
              end
            | "join" -> 
              begin
                debug (sprintf_msg msg);
                let session = msg.session in
                let role = msg.srole in
                debug ("Receiving join message from client") ;
                if Hashtbl.mem invitationtable (session,role)
                then ( (* Create the queues + notification queue *) 
                  let rl,g = Hashtbl.find invitationtable (session,role) in
                  let localtype = project g role in
                  let node_table = build_node_table localtype in
                  let lq = ("Adm",Queue.create ())::
                    (List.map (fun (r,_,_)-> (r,Queue.create())) rl) in
                  Hashtbl.add roletable (session,role) 
                    {lt=localtype;nt=node_table;lq=lq};
                  forward_server (mdeliver,mcheader,mcontent)
                )
                else (
                  debug ("Cannot join if no invitation has been received ...");
                );
                
                ()
              end
            | _ -> (
              debug (Printf.sprintf "Putting %s in the queue of session %s"
                       (sprintf_msg msg) msg.session);
              let state = Hashtbl.find roletable (msg.session,msg.srole) in
              let q = List.assoc msg.srole state.lq in
              Queue.push (Client (mdeliver,mcheader,mcontent,msg)) q;
              Mutex.lock notification_lock;
              Queue.push (msg.session,msg.srole,q) notification;
              Mutex.unlock notification_lock;
              Condition.signal notification_condition
            ) (* Put in the right queue + notification queue *)
          )
        in
        mytreat (mdeliver,mcheader,mcontent,m);
        Queue.iter mytreat client_temp;
        Queue.clear  client_temp;
        Mutex.unlock monitor_lock
      end
    done
  with Queue.Empty -> (debug ("Closing the client thread ...") ;)

let rec check state (v:tempmsg) = function 
  | LEnd n -> None
  | LStop n -> None
  | LGoto n -> (
    match Hashtbl.find state.nt n with
        None -> assert false
      | Some lt -> check state v lt)
  | LSend (n,((r,rr),ml)) -> begin
      match v with
        | Client (mdeliver,mcheader,mcontent,msg) ->
            check_msg msg ml
        | _ -> None
    end
  | LRecv (n,((s,ss),ml)) ->  begin
      match v with
          Server (mdeliver,mcheader,mcontent,msg) ->
            check_msg msg ml
        | _ -> None
    end
  | LPoll (n,((x,r),pl,g,c)) ->  ( assert false  )
  | LPar (n,(LEnd _ , LEnd _,c)) -> check state v c
  | LPar (n,(g,r,c)) -> (
    match check state v g with
      | Some lt -> Some (LPar (n,(lt,r,c)))
      | None -> (
        match check state v r with
          | Some lt -> Some (LPar (n,(g,lt,c)))
          | None -> None))
  | LMu (n,g) -> check state v g
  | LStart (n,g) -> begin
      match v with
          Server (mdeliver,mcheader,mcontent,msg) 
            when msg.lbl = "reg" -> Some g
        | _ -> None
    end
  | LInv (n,g) -> check state v g
  | LRec (n,g) -> check state v g
and check_msg (msg:session_message) = function
    [] -> None
  | ((l,t),lv,lt)::q when (l = msg.lbl (* and the list of values *)) -> Some lt
  | (_)::q -> check_msg msg q
    

let monitor () =
  try 
    while !monitoring do
      Mutex.lock notification_lock;
      if Queue.is_empty notification then
        Condition.wait notification_condition notification_lock;
      debug (Printf.sprintf "Monitoring a message");
      let (session,role,q) = Queue.pop notification in
      debug (Printf.sprintf "There are %d messages left to check" (Queue.length notification));
      Mutex.unlock notification_lock;
      (*    if Queue.is_empty notification 
            then (Thread.yield ())
            else *)
      begin
        Mutex.lock monitor_lock;
        (*      debug (Printf.sprintf "Monitoring a message");
                let (session,role,q) = Queue.pop notification in*)
        let state = Hashtbl.find roletable (session,role) in
        if role = "Adm" || role = "" then begin
          match Queue.pop q with
              Client (mdeliver,mcheader,mcontent,msg) ->
                forward_server (mdeliver,mcheader,mcontent)
            | Server (mdeliver,mcheader,mcontent,msg) ->
              forward_client (mdeliver,mcheader,mcontent)
        end
        else begin
          let v = Queue.pop q in
          (match check state v state.lt with
            | Some lt -> (
              debug (Printf.sprintf "Updating the session state");
              state.lt <- lt ;
              match v with
                  Client (mdeliver,mcheader,mcontent,msg) ->
                    forward_server (mdeliver,mcheader,mcontent)
                | Server (mdeliver,mcheader,mcontent,msg) ->
                  forward_client (mdeliver,mcheader,mcontent))
            | None -> (debug ("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n ... Ignoring illegal message !!!!! \n
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")))
        end;
        Mutex.unlock monitor_lock
      end
    done
  with Queue.Empty -> (debug ("Closing the monitor ...") ;)


(******************)
(* Main procedure *)
(******************)

let main () = 
  debug (Printf.sprintf "Starting the monitor");
  let client_port =
    if Array.length (Sys.argv) = 2 
    then (debug (Printf.sprintf "Reading arguments");
          int_of_string (Sys.argv.(1)))
    else 5673
  in
  debug (Printf.sprintf "Waiting for the client");
  let mutex_start = Mutex.create () in
  let condition_start = Condition.create () in
  Mutex.lock mutex_start;
  let listening_thread =
    Tcp.create_tcp_server ~condition:(Some condition_start) ~max:(Some 1) ~address:"127.0.0.1" ~port:client_port () in
(*  while !Tcp.out_channel = [] && !Tcp.in_descr = [] do
    Thread.yield ()
    done; *)
  Condition.wait condition_start mutex_start;
  Mutex.unlock mutex_start;
  debug (Printf.sprintf "The client connects");
  client_out := snd (List.hd !Tcp.out_channel);
  client_in := List.hd !Tcp.in_descr;
  debug (Printf.sprintf "The client connects");
  (* Protocol Header *)
  let hello = String.create 8 in
  let _ = Unix.recv !client_in hello 0 8 [] in
  assert (hello = Printf.sprintf "AMQP\000\000\009\001");
  debug (Printf.sprintf "Hello from the client");
  debug (Printf.sprintf "Connection to the server");
  let (outc,s) = Amqp.open_tcp_broker () in
  Amqp.out_amqp := outc;
  Amqp.in_amqp := s;
  Amqp.send_hello ();
  let m,_ = Amqp.recv_amqp_constrained 
    ["Connection","Start"] in
  send_client m;
  let m,_ = Amqp.recv_amqp_constrained 
    ~in_amqp:client_in ~messages:client_messages
    ["Connection","Start_Ok"] in
  send_amqp m;
  let m,_ = Amqp.recv_amqp_constrained 
    ["Connection","Tune"] in
  send_client m;
  let m,_ = Amqp.recv_amqp_constrained 
    ~in_amqp:client_in ~messages:client_messages
    ["Connection","Tune_Ok"] in
  send_amqp m;
  let m,_ = Amqp.recv_amqp_constrained 
    ~in_amqp:client_in ~messages:client_messages
    ["Connection","Open"] in
  send_amqp m;
  let m,_ = Amqp.recv_amqp_constrained 
    ["Connection","Open_Ok"] in
  send_client m;
  let m,_ = Amqp.recv_amqp_constrained 
    ~in_amqp:client_in ~messages:client_messages
    ["Channel","Open"] in
  send_amqp m;
  let m,_ = Amqp.recv_amqp_constrained 
    ["Channel","Open_Ok"] in
  send_client m;
  let m,_ = Amqp.recv_amqp_constrained 
    ~in_amqp:client_in ~messages:client_messages
    ["Queue","Declare"] in
  send_amqp m;
  let m,_ = Amqp.recv_amqp_constrained 
    ["Queue","Declare_Ok"] in
  send_client m;
  let m,_ = Amqp.recv_amqp_constrained 
    ~in_amqp:client_in ~messages:client_messages
    ["Basic","Consume"] in
  send_amqp m;
  let m,_ = Amqp.recv_amqp_constrained 
    ["Basic","Consume_Ok"] in
  send_client m;



  debug (Printf.sprintf "Listening to the server");
  let t = Thread.create listening_server () in
  listening_server_thread := Some t;
  
  debug (Printf.sprintf "Listening to the client");
  let t = Thread.create listening_client () in
  listening_client_thread := Some t;

  debug (Printf.sprintf "Treating the server messages");
  let t = Thread.create treat_server () in
  treating_server_thread := Some t;

  debug (Printf.sprintf "Treating the client messages");
  let t = Thread.create treat_client () in
  treating_client_thread := Some t;

  debug (Printf.sprintf "Monitoring all messages");
  let t = Thread.create monitor () in
  monitor_thread := Some t;
  (* The monitor does not stop *)
  Thread.join (Common.get_option !listening_server_thread);
  debug (Printf.sprintf "Closing 1");
  Thread.join (Common.get_option !listening_client_thread);
  debug (Printf.sprintf "Closing 2");
  Thread.join (Common.get_option !treating_server_thread);
  debug (Printf.sprintf "Closing 3");
  Thread.join (Common.get_option !treating_server_thread);
  debug (Printf.sprintf "Closing 4");
  Condition.signal notification_condition;
  Thread.join (Common.get_option !monitor_thread);
  debug (Printf.sprintf "Closing 5");
  Thread.join (listening_thread);
  debug (Printf.sprintf "Closing!!!!!!");
  ()

    

let () = main ()
